/* See bmpcount/doc/COPYING for copyright/licensing information. */
/* cestan@cs.ucsd.edu 10/22/2001 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "adaptivebmp.h"
#include "../readwrite/assume.h"

static const double rhomax = 2.6744;
static const double bitsetratio = 0.93105; //(1-1/exp(rhomax))

abmp::abmp(int bb,int cc,int bigbb,int spanbigg,int memory,double rhominbig,
	   double rhomaxbig,int flowIDsize):hash(flowIDsize){
  int i;
  b=bb;
  c=cc;
  bigb=bigbb;
  spanbig=spanbigg;
  assume(b>1,"Components must have at least two bits\n");
  assume(bigb>=b,"The big component must be larger than the others\n");
  assume(spanbig>0,"Spanbig must be positive\n");
  assume(c>spanbig,"There have to be more components than spanbig\n");
  lastb=(memory>0?memory-(c-1-spanbig)*b-bigb:b*2);
  assume2(lastb>=b*2,"Not enough memory for last component: lastb=%d b=%d\n",
	  lastb,b);
  assume2(log(b)+log(2)*c<32*log(2),
	  "Adaptive bitmap too fine for 32 bit hash: bigb=%d c=%d\n",bigb,c);
  if(log(b)+log(2)*c>27*log(2)){
    fprintf(stderr,"Unsafe abmp for a 32 bit hash: bigb=%d c=%d\n",bigb,c);
  }
  assume(rhominbig<=rhomax/2,"Rhominbig must be below 1.3372\n");
  assume(rhomaxbig>=rhomax,"Rhomaxbig must be below 2.6744\n");
  setmax=(int)(b*bitsetratio+0.5);
  bigsetmin=(int)(bigb*(1-1/exp(rhominbig))+0.5);
  bigsetmax=(int)(bigb*(1-1/exp(rhomaxbig))+0.5);
  posbig=0;
  bigoffset=0;
  offsets[0]=0;
  offsets[1]=bigb;
  for(i=2;i<spanbig;i++){
    offsets[i]=-0xBADBAD;
  }
  for(i=spanbig;i<c;i++){
    offsets[i]=bigb+(i-spanbig)*b;
  }
  offsets[c]=offsets[c-1]+lastb;
  bmp=allocbitmap(offsets[c]);
  assumenotnull(bmp);
  fillzero(bmp,offsets[c]);
}

abmp::~abmp(){
  free(bmp);
}


void abmp::new_packet(byte* flowID){
  uint32bit key=hash.compute(flowID);
  int pos,crt;
  for(crt=0;crt<c-1;crt++){
    if(!(key&1)){/*this is the component*/
      key=key>>1;
      break;
    } else {
      key=key>>1;
    }
  }
  if(crt>=posbig && crt<posbig+spanbig){
    /* we map to the big virtual bitmap */
    pos=key%bigb;
    setbit(pos+bigoffset,bmp);
    return;
  }
  if(crt<c-1){
    pos=key%b;
    setbit(pos+offsets[crt],bmp);
  } else {
    pos=key%lastb;
    setbit(pos+offsets[c-1],bmp);
  }
}

int abmp::estimate(){
  int base,i,factor,z,j;
  double m;
  factor=1;
  // use big component if it has between bigsetmin and bigsetmax bits set
  z=countzerobits(bmp,bigoffset,bigoffset+bigb);
  if((bigb-z<=bigsetmax && bigb-z>=bigsetmin) ||
     (bigb-z<=bigsetmax && posbig==0) ||
     (bigb-z>=bigsetmin && posbig==c-spanbig)){
    base=posbig;
    for(i=0;i<posbig;i++){
      factor=factor*2;
    }
  } else {
    // Choose the base from among the normal components.
    fprintf(stderr,"Base not big virtual bitmap!!!\n");
    if(bigb-z>bigsetmax){
      base=posbig+spanbig;
      for(i=0;i<posbig+spanbig;i++){
	factor=factor*2;
      }
    } else {
      base=0;
    }
    for(;base<c-1;base++){
      assume(base!=posbig,"Adaptive bitmap anomaly\n");
      z=countzerobits(bmp,offsets[base],offsets[base]+b);
      if(b-z<=setmax){
	break;
      } else {
	factor=factor*2;
      }
    }
  }
  /* Estimate the number of flows from the base up*/
  m=0;
  for(i=base;i<c;i++){
    z=countzerobits(bmp,offsets[i],offsets[i+1]);
    if(z==0){
      if(i==c-1 || (i==posbig && posbig==c-spanbig)){
	return -1; // Finest component totally full
      }
      assume4(z!=0,"Bug or statistical anomaly: full component found after base\n crt=%d base=%d posbig=%d spanbig=%d \n",i,base,posbig,spanbig);
    }
    m+=(offsets[i+1]-offsets[i])*(log(offsets[i+1]-offsets[i])-log(z));
    if(i==posbig){
      i+=spanbig-1;
    }
  }
  return (int)(factor*m+0.5);
}

static const double magicrhomax=1.5949; // Best accuracy here

void abmp::reset(){
  int i;
  double maxaccurate=bigb*magicrhomax;
  double ratio = estimate()/maxaccurate;
  if(ratio<1){
    posbig=0;
  } else {
    posbig=(int)(log(ratio)/log(2)+0.5);
    if(posbig>c-spanbig){/* cannot leave components uncovered */
      posbig=c-spanbig;
    }
  }
  for(i=1;i<=posbig;i++){
    offsets[i]=offsets[i-1]+b;
  }
  offsets[posbig+1]=offsets[posbig]+bigb;
  for(i=posbig+2;i<posbig+spanbig;i++){
    offsets[i]=-0xBADBAD;
  }
  offsets[posbig+spanbig]=offsets[posbig+1];
  for(i=posbig+spanbig+1;i<c;i++){
    offsets[i]=offsets[i-1]+b;
  }
  if(posbig+spanbig<c){
    offsets[c]=offsets[c-1]+lastb;
  }
  fillzero(bmp,offsets[c]);
}

