/* See bmpcount/doc/COPYING for copyright/licensing information. */
/* cestan@cs.ucsd.edu 07/12/2002 */

/* The class hierarchy rooted in the "thing" class has derived classes
   such as: an IP address, a port range, a flow identifier 5-tuple,
   etc. By relying on the inheritance relations from this hierarchy we
   can implement nice generic algorithms. There is a separation
   between data and functionality so that we can efficiently implement
   things such as flow lookup tables without having thousands or millions
   of objects (objects for each table key). Instead we have a single
   flow ID object (containing IP address and port number objects) that
   handles all the instances of flow IDs in the table. In general each
   "thing" object can be associated with an arbitrary number of
   instances of data that use the same amount of memory each. The
   correspondence between the object and the data is maintained by
   external objects (such as the hash table implementation). All
   things also have a human readable (text) form. It is used in data
   files (that get gzipped anyway) and for output that is supposed to
   be read by humans. All things must know how to convert between the
   human readable and internal (binary) form of their data. */

#ifndef THING_H
#define THING_H

#include "../readwrite/buffer.h"

class thing{
 public:
  virtual unsigned int getbinsize()=0;
  /* Size of binary representation (in bytes) */
  virtual unsigned int getmaxhrsize()=0;
  /* Maximum size of human readable form of this thing */
  virtual void print (buffer& b, const  byte* data)=0;
  /* Prints the current value of the thing */
  virtual int parse (buffer& b, byte* data)=0;
  /* Parses from the buffer -- return 1 on success and 0 on failure */
  virtual void printname(buffer& b)=0;
  /* Prints the name of the thing (e.g. Source IP) */

  virtual void prettyprint(buffer& b, const byte* data)=0;
  /* Prints the value of the thing with padding up to maxhrsize + 1 space*/
  virtual void prettyprintname(buffer& b)=0;
  void debugprint(const byte* data); 
  /* Prints data onto stderr (for debugging) */
};

#endif /* THING_H */
