#!/usr/bin/perl

# This script computes the configuration of the multiresolution bitmap
# using the minimum number of bits that achieves the required accuracy
# for up to the given number of flows.

# It has to receive on the standard input the output of
# findconstantsformrbmp.pl

sub initialize_constants(){
  while (<>){
    my (@F)=split(/\s+/);
    $rhomin{$F[0]}=$F[1];
    $rhomax{$F[0]}=$F[2];
    $coef{$F[0]}=$F[3];
  }
}

# Computes the error of a virtual bitmap 
sub vbmperror($$){
  my ($rho,$b)=($_[0],$_[1]);
  return (sqrt(exp($rho)-1)/$rho/sqrt($b));
}

# Computes the error of the multiresolution bitmap component before last

sub mrbblerror($$$$){
  my ($rho,$k,$b,$blast)=($_[0],$_[1],$_[2],$_[3]);
  return (sqrt($b*(exp($rho)-1)+$blast*(exp($rho*$b/($k-1)/$blast)-1))*($k-1)/$k/$rho/$b);
}

# Puts into $b, $blast and $c the configuration of the smallest mrb
# good enough for $n with $epsilon
sub computeconfiguration($$$$){
  my ($epsilon,$N,$k,$hardware)=($_[0],$_[1],$_[2],$_[3]);

  # This configuration works for sure.
  $b=int($coef{$k}/($epsilon*$epsilon)+0.999999999);
  $blast=int($coef{$k}/($epsilon*$epsilon)*$k/($k-1)+0.999999999);
  $rhomax=$rhomax{$k};
  $rhomin=$rhomin{$k};
  $c=2+int(log($N/($b*$rhomax))/log($k)+0.999999999);
  if(not $hardware){
    # But we need to adjust $blast so that the component before last works
    while(mrbblerror($rhomin,$k,$b,$blast)>$epsilon or mrbblerror($rhomax,$k,$b,$blast)>$epsilon){
      $blast++;
    }
    # But maybe increasing the size of the last component can lead to
    # using fewer components and maybe to smaller total memory.
    #
    # First we extend it enough so that it can take it over from the
    # last normal component.
    $biggerblast=$blast;
    while(vbmperror($rhomin*$blast/$biggerblast,$biggerblast)>$epsilon){
      $biggerblast+=1;
    }
    $rhomaxlast=1.6;
    while(vbmperror($rhomaxlast,$biggerblast)<$epsilon){
      $rhomaxlast+=0.00002;
    }
    $rhomaxlast-=0.00002;
    $cbigger=1+int(log($N/($biggerblast*$rhomaxlast))/log($k)+0.999999999);
    # And now we see how much more we would need to extend it to get
    # rid of one more component.
    $biggest=$biggerblast;
    while(1+int(log($N/($biggest*$rhomaxlast))/log($k)+0.999999999)==$cbigger){
      $biggest++;
      while(vbmperror($rhomaxlast,$biggest)<$epsilon){
	$rhomaxlast+=0.00002;
      }
      $rhomaxlast-=0.00002;
    }
  } else {
    # hardware case -- we have to work with powers of two
    #print "b $b->";
    $blastold=$blast;
    $blast=2**(int(log($blast)/log(2)+0.999999999));
    $b=int($blast*($k-1)/$k+0.000001);
    #print "$b\tblast $blastold->$blast\n";
    # But we need to adjust $blast so that the component before last works
    while(mrbblerror($rhomin,$k,$b,$blast)>$epsilon or mrbblerror($rhomax,$k,$b,$blast)>$epsilon){
      $blast=$blast*2;
    }
    # But maybe increasing the size of the last component can lead to
    # using fewer components and maybe to smaller total memory.
    #
    # First we extend it enough so that it can take it over from the
    # last normal component.
    $biggerblast=$blast;
    while(vbmperror($rhomin*$blastold/$biggerblast,$biggerblast)>$epsilon){
      $biggerblast=$biggerblast*2;
    }
    $rhomaxlast=1.6;
    while(vbmperror($rhomaxlast,$biggerblast)<$epsilon){
      $rhomaxlast+=0.00002;
    }
    $rhomaxlast-=0.00002;
    $cbigger=1+int(log($N/($biggerblast*$rhomaxlast))/log($k)+0.999999999);
    # And now we see how much more we would need to extend it to get
    # rid of one more component.
    $biggest=$biggerblast;
    while(1+int(log($N/($biggest*$rhomaxlast))/log($k)+0.999999999)==$cbigger){
      $biggest=$biggest*2;
      while(vbmperror($rhomaxlast,$biggest)<$epsilon){
	$rhomaxlast+=0.00002;
      }
      $rhomaxlast-=0.00002;
    }
  }
  if($c*$b+$blast-$b>$cbigger*$b+$biggerblast-$b){
    $c=$cbigger;
    $blast=$biggerblast;
  }
  if($c*$b+$blast-$b>$cbigger*$b+$biggest-2*$b){
    $c=$cbigger-1;
    $blast=$biggest;
  }
}

initialize_constants();

print "***** Software configurations *****\n";
foreach $epsilon (0.1,0.03,0.01){
  foreach $k (2,3,4,5,6,7,8){
    foreach $N (1000000,100000000,10000000000){
      print "e=$epsilon\tk=$k\tN=$N\t";
      computeconfiguration($epsilon,$N,$k,0);
      print "b=$b\tc=$c\tblast=$blast\tmem=".($c*$b+$blast-$b)."\n";
    }
    print "-------------------------------------------------\n";
  }
}
print "***** Hardware configurations *****\n";
foreach $epsilon (0.1,0.03,0.01){
  foreach $k (2,4,8){
    foreach $N (1000000,100000000,10000000000){
      print "e=$epsilon\tk=$k\tN=$N\t";
      computeconfiguration($epsilon,$N,$k,1);
      print "b=$b\tc=$c\tblast=$blast\tmem=".($c*$b+$blast-$b)."\n";
    }
    print "-------------------------------------------------\n";
  }
}
